use std::{borrow::Borrow, ops::Deref};

use crate::{ValidationFailReason, ValidationResult};

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
#[repr(transparent)]
pub struct Email(pub(crate) String);

impl Email {
    /// Emails must contain exactly one @ and have characters either side.
    pub fn new(post_code: impl Into<String>) -> ValidationResult<Self> {
        let post_code = post_code.into();

        validate_email(&post_code)?;

        Ok(Self(post_code))
    }

    pub fn as_str(&self) -> &str {
        &self.0
    }
}

impl Borrow<EmailRef> for Email {
    fn borrow(&self) -> &EmailRef {
        let value = self.0.as_str();

        // Safety:
        // EmailRef is a wrapper of str, therefore converting a &str to a &EmailRef is safe.
        unsafe { &*(value as *const _ as *const EmailRef) }
    }
}

impl AsRef<EmailRef> for Email {
    fn as_ref(&self) -> &EmailRef {
        self.borrow()
    }
}

impl Deref for Email {
    type Target = EmailRef;

    fn deref(&self) -> &Self::Target {
        self.borrow()
    }
}

impl PartialEq<EmailRef> for Email {
    fn eq(&self, other: &EmailRef) -> bool {
        self.0.eq(&other.0)
    }
}

#[derive(Debug, PartialEq, Eq, Hash)]
#[repr(transparent)]
pub struct EmailRef(str);

impl EmailRef {
    /// Emails must contain exactly one @ and have characters either side.
    pub fn new(post_code: &str) -> ValidationResult<&Self> {
        validate_email(&post_code)?;

        // Safety:
        // EmailRef is a wrapper of str, therefore converting a &str to a &EmailRef is
        // safe.
        Ok(unsafe { &*(post_code as *const _ as *const Self) })
    }

    pub fn as_str(&self) -> &str {
        &self.0
    }
}

impl ToOwned for EmailRef {
    type Owned = Email;

    fn to_owned(&self) -> Self::Owned {
        Email(self.0.to_owned())
    }
}

impl PartialEq<Email> for EmailRef {
    fn eq(&self, other: &Email) -> bool {
        self.0.eq(&other.0)
    }
}

// TODO: This should probably be less shit
fn validate_email(email: &str) -> ValidationResult<()> {
    let mut part_count = 0;

    for part in email.split('@') {
        if part.is_empty() {
            Err(ValidationFailReason::Email)?;
        }

        part_count += 1;
    }

    if part_count != 2 {
        Err(ValidationFailReason::Email)?;
    }

    Ok(())
}
