//! Input validation module
mod email;
mod password;
mod phone_number;
mod post_code;

use std::fmt;

#[cfg(feature = "serialize")]
use serde::Serialize;

pub use crate::{email::*, password::*, phone_number::*, post_code::*};

/// Which address line are we talking about? Who knows?
#[derive(Debug, Clone, PartialEq, Eq)]
#[cfg_attr(feature = "serialize", derive(Serialize))]
pub enum AddressLine {
    One,
    Two,
    City,
}

impl fmt::Display for AddressLine {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Self::One => write!(f, "Address Line One"),
            Self::Two => write!(f, "Address Line Two"),
            Self::City => write!(f, "City"),
        }
    }
}

/// Reason for a validation failing.
#[derive(thiserror::Error, Debug, Clone, PartialEq)]
#[cfg_attr(feature = "serialize", derive(Serialize))]
pub enum ValidationFailReason {
    #[error("Names must be longer than one character")]
    Name,
    #[error("{0} must be more than one character long")]
    AddressLine(AddressLine),
    #[error("Please enter a valid post code. Including the space in the middle.")]
    PostCode,
    #[error("Please enter a valid email address.")]
    Email,
    #[error("Passwords must have an entropy value of at least {0}. A password's entropy is a measure of strength. Entropy is increased with the length of the password and the size of the alphabet used to construct it.")]
    Password(f64),
    #[error("Please provide a value.")]
    EmptyString,
    #[error("{0}")]
    PhoneNumber(#[from] PhoneNumberError),
}

pub type ValidationResult<T> = Result<T, ValidationFailReason>;

/// Validate that the string value is not empty.
pub fn validate_non_empty_string(test: &str) -> Result<(), ValidationFailReason> {
    if test.trim().len() > 0 {
        Ok(())
    } else {
        Err(ValidationFailReason::EmptyString)
    }
}

/// Check if a string is a valid name.
/// Must be more than one char long
pub fn is_valid_name(test: &str) -> bool {
    !test.is_empty()
}

/// Check if a string is a valid name.
/// Must be more than one char long
pub fn validate_name(test: &str) -> Result<(), ValidationFailReason> {
    if is_valid_name(test) {
        Ok(())
    } else {
        Err(ValidationFailReason::Name)
    }
}

/// Must be more that one char long
pub fn is_valid_address_line(test: &str) -> bool {
    !test.is_empty()
}

/// Validate an email address, returning a result.
pub fn validate_address_line(
    test: &str,
    address_line: AddressLine,
) -> Result<(), ValidationFailReason> {
    if is_valid_address_line(test) {
        Ok(())
    } else {
        Err(ValidationFailReason::AddressLine(address_line))
    }
}

#[cfg(test)]
mod test {
    macro_rules! simple_test {
        ($name:ident, $expected:expr, $test:expr, $fn:tt) => {
            #[test]
            fn $name() {
                assert_eq!($expected, $fn($test));
            }
        };
    }

    use super::{
        is_valid_address_line,
        is_valid_name,
        validate_non_empty_string,
        Email,
        Password,
        PasswordStrength,
        PhoneNumber,
        PhoneNumberError,
        PostCode,
        ValidationFailReason,
    };

    // validate_non_empty_string
    simple_test!(
        validate_non_empty_string_1,
        Ok(()),
        "Hello",
        validate_non_empty_string
    );

    simple_test!(
        validate_non_empty_string_2,
        Err(ValidationFailReason::EmptyString),
        " ",
        validate_non_empty_string
    );

    simple_test!(
        validate_non_empty_string_3,
        Err(ValidationFailReason::EmptyString),
        "  ",
        validate_non_empty_string
    );

    simple_test!(
        validate_non_empty_string_4,
        Err(ValidationFailReason::EmptyString),
        "",
        validate_non_empty_string
    );

    // is_valid_name
    simple_test!(is_valid_name_1, true, "Michael", is_valid_name);
    simple_test!(is_valid_name_2, true, "m", is_valid_name);
    simple_test!(is_valid_name_3, false, "", is_valid_name);

    // is_valid_phone_number
    simple_test!(
        is_valid_phone_number_1,
        Ok(PhoneNumber(*b"07858298563")),
        "07858298563",
        (PhoneNumber::new)
    );
    simple_test!(
        is_valid_phone_number_2,
        Err(ValidationFailReason::PhoneNumber(
            PhoneNumberError::InvalidFirstCharacter
        )),
        "jjjjjjjjjjj",
        (PhoneNumber::new)
    );
    simple_test!(
        is_valid_phone_number_3,
        Err(ValidationFailReason::PhoneNumber(
            PhoneNumberError::InvalidLength
        )),
        "0785829856",
        (PhoneNumber::new)
    );
    simple_test!(
        is_valid_phone_number_4,
        Err(ValidationFailReason::PhoneNumber(
            PhoneNumberError::InvalidFirstCharacter
        )),
        "78582985633",
        (PhoneNumber::new)
    );
    simple_test!(
        is_valid_phone_number_5,
        Err(ValidationFailReason::PhoneNumber(
            PhoneNumberError::InvalidFirstCharacter
        )),
        "...........",
        (PhoneNumber::new)
    );
    simple_test!(
        is_valid_phone_number_6,
        Err(ValidationFailReason::PhoneNumber(
            PhoneNumberError::InvalidCharacter(10, b'.')
        )),
        "0785829856.",
        (PhoneNumber::new)
    );
    simple_test!(
        is_valid_phone_number_7,
        Err(ValidationFailReason::PhoneNumber(
            PhoneNumberError::InvalidCharacter(10, b'k')
        )),
        "0785829856k",
        (PhoneNumber::new)
    );

    // is_valid_address_line
    simple_test!(
        is_valid_address_line_1,
        true,
        "1 proctors way",
        is_valid_address_line
    );
    simple_test!(
        is_valid_address_line_2,
        true,
        "hello there",
        is_valid_address_line
    );
    simple_test!(is_valid_address_line_3, true, "m", is_valid_address_line);
    simple_test!(is_valid_address_line_4, false, "", is_valid_address_line);

    // is_valid_postcode
    simple_test!(
        is_valid_postcode_1,
        Ok(PostCode("dn20 9qf".to_owned())),
        "dn20 9qf",
        (PostCode::new)
    );
    simple_test!(
        is_valid_postcode_2,
        Ok(PostCode("dn20 9qf".to_owned())),
        "dn20 9qf",
        (PostCode::new)
    );
    simple_test!(
        is_valid_postcode_3,
        Ok(PostCode("DN20 9qf".to_owned())),
        "DN20 9qf",
        (PostCode::new)
    );
    simple_test!(
        is_valid_postcode_4,
        Ok(PostCode("dn20 9QF".to_owned())),
        "dn20 9QF",
        (PostCode::new)
    );
    simple_test!(
        is_valid_postcode_5,
        Ok(PostCode("w1 9qf".to_owned())),
        "w1 9qf",
        (PostCode::new)
    );
    simple_test!(
        is_valid_postcode_6,
        Ok(PostCode("s1w 9qf".to_owned())),
        "s1w 9qf",
        (PostCode::new)
    );
    simple_test!(
        is_valid_postcode_7,
        Err(ValidationFailReason::PostCode),
        "1ws 9qf",
        (PostCode::new)
    );
    simple_test!(
        is_valid_postcode_8,
        Err(ValidationFailReason::PostCode),
        "1w 9qf",
        (PostCode::new)
    );

    // is_valid_email
    simple_test!(
        is_valid_email_1,
        Ok(Email("bruh@me.com".to_owned())),
        "bruh@me.com",
        (Email::new)
    );
    simple_test!(
        is_valid_email_2,
        Err(ValidationFailReason::Email),
        "7878787878",
        (Email::new)
    );
    simple_test!(
        is_valid_email_3,
        Err(ValidationFailReason::Email),
        "bruh",
        (Email::new)
    );
    simple_test!(
        is_valid_email_4,
        Err(ValidationFailReason::Email),
        "",
        (Email::new)
    );
    simple_test!(
        is_valid_email_5,
        Err(ValidationFailReason::Email),
        "878",
        (Email::new)
    );
    simple_test!(
        is_valid_email_6,
        Err(ValidationFailReason::Email),
        "BRUH",
        (Email::new)
    );

    // is_valid_password
    simple_test!(
        is_valid_password_1,
        (
            Password("HelloThere123".to_owned()),
            PasswordStrength(77.40455203502937)
        ),
        "HelloThere123",
        (Password::new)
    );
    simple_test!(
        is_valid_password_2,
        (
            Password("hellothere.12".to_owned()),
            PasswordStrength(79.13701693625441)
        ),
        "hellothere.12",
        (Password::new)
    );
    simple_test!(
        is_valid_password_3,
        (
            Password("HELLOTHERE.1".to_owned()),
            PasswordStrength(73.04955409500407)
        ),
        "HELLOTHERE.1",
        (Password::new)
    );
    simple_test!(
        is_valid_password_4,
        (
            Password("HelloThere.".to_owned()),
            PasswordStrength(70.31549165056637)
        ),
        "HelloThere.",
        (Password::new)
    );
    simple_test!(
        is_valid_password_5,
        (
            Password("Hh1".to_owned()),
            PasswordStrength(17.862588931160623)
        ),
        "Hh1",
        (Password::new)
    );
    simple_test!(
        is_valid_password_6,
        (
            Password("hellothere1".to_owned()),
            PasswordStrength(56.86917501586544)
        ),
        "hellothere1",
        (Password::new)
    );
    simple_test!(
        is_valid_password_7,
        (
            Password("HELLOTHERE1".to_owned()),
            PasswordStrength(56.86917501586544)
        ),
        "HELLOTHERE1",
        (Password::new)
    );
    simple_test!(
        is_valid_password_8,
        (
            Password("hellothere.".to_owned()),
            PasswordStrength(64.4377909464033)
        ),
        "hellothere.",
        (Password::new)
    );
    simple_test!(
        is_valid_password_9,
        (
            Password("HELLOTHERE.".to_owned()),
            PasswordStrength(64.4377909464033)
        ),
        "HELLOTHERE.",
        (Password::new)
    );
    simple_test!(
        is_valid_password_10,
        (Password("".to_owned()), PasswordStrength(0.0)),
        "",
        (Password::new)
    );
}
