use std::{
    borrow::Borrow,
    ops::{Deref, Range, RangeFrom},
};

#[derive(Debug, Clone, Copy, PartialEq, PartialOrd)]
pub struct PasswordStrength(pub(crate) f64);

impl PasswordStrength {
    pub fn is_weak(self) -> bool {
        const POOR_RANGE: Range<f64> = 0.0..65.0;

        POOR_RANGE.contains(&self.0)
    }

    pub fn is_reasonable(self) -> bool {
        const POOR_RANGE: Range<f64> = 65.0..100.0;

        POOR_RANGE.contains(&self.0)
    }

    pub fn is_strong(self) -> bool {
        const POOR_RANGE: RangeFrom<f64> = 100.0..;

        POOR_RANGE.contains(&self.0)
    }
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
#[repr(transparent)]
pub struct Password(pub(crate) String);

impl Password {
    /// Returns a new password and it's strength as a tuple. There is no requirement on strength
    /// enforced here, it is up to the application to decide how strong a password should be.
    pub fn new(password: impl Into<String>) -> (Self, PasswordStrength) {
        let password = Self(password.into());
        let strength = PasswordStrength(calculate_entropy(&password.0));

        (password, strength)
    }

    /// Recalculates the strength of the password contained within self. This means you don't need
    /// to store the strength along side the password.
    pub fn strength(&self) -> PasswordStrength {
        PasswordStrength(calculate_entropy(&self.0))
    }

    pub fn as_str(&self) -> &str {
        &self.0
    }
}

impl Borrow<PasswordRef> for Password {
    fn borrow(&self) -> &PasswordRef {
        let value = self.0.as_str();

        // Safety:
        // PasswordRef is a wrapper of str, therefore converting a &str to a &PasswordRef is safe.
        unsafe { &*(value as *const str as *const PasswordRef) }
    }
}

impl AsRef<PasswordRef> for Password {
    fn as_ref(&self) -> &PasswordRef {
        self.borrow()
    }
}

impl Deref for Password {
    type Target = PasswordRef;

    fn deref(&self) -> &Self::Target {
        self.borrow()
    }
}

impl PartialEq<PasswordRef> for Password {
    fn eq(&self, other: &PasswordRef) -> bool {
        self.0.eq(&other.0)
    }
}

#[derive(Debug, PartialEq, Eq, Hash)]
#[repr(transparent)]
pub struct PasswordRef(pub(crate) str);

impl PasswordRef {
    /// Returns a new password and it's strength as a tuple. There is no requirement on strength
    /// enforced here, it is up to the application to decide how strong a password should be.
    pub fn new(password: &str) -> (&Self, PasswordStrength) {
        (
            // Safety:
            // PasswordRef is a wrapper of str, therefore converting a &str to a &PasswordRef is
            // safe.
            unsafe { &*(password as *const _ as *const Self) },
            PasswordStrength(calculate_entropy(password)),
        )
    }

    /// Recalculates the strength of the password contained within self. This means you don't need
    /// to store the strength along side the password.
    pub fn strength(&self) -> PasswordStrength {
        PasswordStrength(calculate_entropy(&self.0))
    }

    pub fn as_str(&self) -> &str {
        &self.0
    }
}

impl ToOwned for PasswordRef {
    type Owned = Password;

    fn to_owned(&self) -> Self::Owned {
        Password(self.0.to_owned())
    }
}

impl PartialEq<Password> for PasswordRef {
    fn eq(&self, other: &Password) -> bool {
        self.0.eq(&other.0)
    }
}

const DIGIT_COUNT: f64 = 10.0;
const LOWER_COUNT: f64 = 26.0;
const UPPER_COUNT: f64 = 26.0;
const PUNCT_COUNT: f64 = 32.0;

/// Calculate the entropy of a password using `E = L * log2(R)` Where L is the length of the
/// password and R is the range of characters to choose from.
///
/// NOTE: This doesn't account for non-ascii passwords.
fn calculate_entropy(password: &str) -> f64 {
    #![allow(non_snake_case)]

    if password.is_empty() {
        0.0
    } else {
        let L = password.len() as f64;
        let R = {
            let mut r_values = [0.0; 4];

            for b in password.bytes() {
                if b.is_ascii_digit() {
                    r_values[0] = DIGIT_COUNT;
                }
                if b.is_ascii_lowercase() {
                    r_values[1] = LOWER_COUNT;
                }
                if b.is_ascii_uppercase() {
                    r_values[2] = UPPER_COUNT;
                }
                if b.is_ascii_punctuation() {
                    r_values[3] = PUNCT_COUNT;
                }
            }

            r_values.iter().fold(0.0, |acc, a| acc + a)
        };

        L * f64::log2(R)
    }
}
