use std::{borrow::Borrow, ops::Deref};

use lazy_static::lazy_static;
use regex::Regex;

use crate::{ValidationFailReason, ValidationResult};

lazy_static! {
    /// Postcode regex.
    /// [Source](https://assets.publishing.service.gov.uk/government/uploads/system/uploads/attachment_data/file/488478/Bulk_Data_Transfer_-_additional_validation_valid_from_12_November_2015.pdf)
    pub static ref POSTCODE_REGEX: Regex = Regex::new("^([Gg][Ii][Rr] 0[Aa]{2})|((([A-Za-z][0-9]{1,2})|(([A-Za-z][A-Ha-hJ-Yj-y][0-9]{1,2})|(([AZa-z][0-9][A-Za-z])|([A-Za-z][A-Ha-hJ-Yj-y][0-9]?[A-Za-z])))) [0-9][A-Za-z]{2})$").unwrap();
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
#[repr(transparent)]
pub struct PostCode(pub(crate) String);

impl PostCode {
    /// Must be longer than one char and match [`struct@POSTCODE_REGEX`]
    pub fn new(post_code: impl Into<String>) -> ValidationResult<Self> {
        let post_code = post_code.into();

        validate_post_code(&post_code)?;

        Ok(Self(post_code))
    }

    pub fn as_str(&self) -> &str {
        &self.0
    }
}

impl Borrow<PostCodeRef> for PostCode {
    fn borrow(&self) -> &PostCodeRef {
        let value = self.0.as_str();

        // Safety:
        // PostCodeRef is a wrapper of str, therefore converting a &str to a &PostCodeRef is safe.
        unsafe { &*(value as *const _ as *const PostCodeRef) }
    }
}

impl AsRef<PostCodeRef> for PostCode {
    fn as_ref(&self) -> &PostCodeRef {
        self.borrow()
    }
}

impl Deref for PostCode {
    type Target = PostCodeRef;

    fn deref(&self) -> &Self::Target {
        self.borrow()
    }
}

impl PartialEq<PostCodeRef> for PostCode {
    fn eq(&self, other: &PostCodeRef) -> bool {
        self.0.eq(&other.0)
    }
}

#[derive(Debug, PartialEq, Eq, Hash)]
#[repr(transparent)]
pub struct PostCodeRef(pub(crate) str);

impl PostCodeRef {
    /// Must be longer than one char and match [`struct@POSTCODE_REGEX`]
    pub fn new(post_code: &str) -> ValidationResult<&Self> {
        validate_post_code(&post_code)?;

        // Safety:
        // PostCodeRef is a wrapper of str, therefore converting a &str to a &PostCodeRef is
        // safe.
        Ok(unsafe { &*(post_code as *const _ as *const Self) })
    }

    pub fn as_str(&self) -> &str {
        &self.0
    }
}

impl ToOwned for PostCodeRef {
    type Owned = PostCode;

    fn to_owned(&self) -> Self::Owned {
        PostCode(self.0.to_owned())
    }
}

impl PartialEq<PostCode> for PostCodeRef {
    fn eq(&self, other: &PostCode) -> bool {
        self.0.eq(&other.0)
    }
}

fn validate_post_code(post_code: &str) -> ValidationResult<()> {
    if !post_code.is_empty() && POSTCODE_REGEX.is_match(&post_code) {
        Ok(())
    } else {
        Err(ValidationFailReason::PostCode)
    }
}
